@@ -21,10 +21,11 @@ class RawTextRfcWriter(BaseRfcWriter):
         The page width is controlled by the *width* parameter.
     """
 
-    def __init__(self, xmlrfc, width=72, quiet=False, verbose=False, date=datetime.date.today()):
+    def __init__(self, xmlrfc, width=72, margin=3, quiet=False, verbose=False, date=datetime.date.today()):
         BaseRfcWriter.__init__(self, xmlrfc, quiet=quiet, verbose=verbose, date=date)
         # Document processing data
         self.width = width      # Page width
+        self.margin = margin    # Page margin
         self.buf = []           # Main buffer during processing
         self.output = []        # Final buffer that gets written to disk
         self.toc_marker = 0     # Line number in buffer to write toc to
@@ -33,7 +34,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         self.edit_counter = 0   # Counter for edit marks
         self.eref_counter = 0   # Counter for <eref> elements
         self.ascii = True       # Enable ascii flag
-        
+
         # Marks position of iref elements
         self.iref_marks = {}
 
@@ -59,8 +60,9 @@ class RawTextRfcWriter(BaseRfcWriter):
         """ <vspace> line break wrapper to allow for overrides """
         return self._lb(num=num)
 
-    def _write_text(self, string, indent=0, sub_indent=0, bullet='', \
-                  align='left', lb=False, buf=None, strip=True, edit=False):
+    def _write_text(self, string, indent=0, sub_indent=0, bullet='',
+                  align='left', leading_blankline=False, buf=None,
+                  strip=True, edit=False, source_line=None):
         """ Writes a line or multiple lines of text to the buffer.
 
             Several parameters are included here.  All of the API calls
@@ -82,7 +84,7 @@ class RawTextRfcWriter(BaseRfcWriter):
             # No bullet, so combine indent and sub_indent
             initial = subsequent
 
-        if lb:
+        if leading_blankline:
             if edit and self.pis.get('editing', 'no') == 'yes':
                 # Render an editing mark
                 self.edit_counter += 1
@@ -100,7 +102,12 @@ class RawTextRfcWriter(BaseRfcWriter):
                 buf.extend(par)
             elif align == 'center':
                 for line in par:
-                    buf.append(line.center(self.width).rstrip())
+                    margin = ' ' * (indent//2*2)
+                    if line.startswith(margin):
+                        centered = margin + line[len(margin):].center(self.width-(indent//2*2)).rstrip()
+                        buf.append(centered)
+                    else:
+                        buf.append(line.center(self.width).rstrip())
             elif align == 'right':
                 for line in par:
                     buf.append(line.rjust(self.width))
@@ -112,7 +119,7 @@ class RawTextRfcWriter(BaseRfcWriter):
     def _write_list(self, list, level=0, indent=3):
         """ Writes a <list> element """
         bullet = '   '
-        hangIndent = 0
+        hangIndent = None
         style = list.attrib.get('style', 'empty')
         if style == 'hanging' or style.startswith('format'):
             # Check for optional hangIndent
@@ -143,11 +150,11 @@ class RawTextRfcWriter(BaseRfcWriter):
                 self.xmlrfc.parse_pi(element)
             elif element.tag == 't':
                 # Disable linebreak if subcompact=yes AND not first list element
-                lb = True
+                leading_blankline = True
                 if t_count > 0 and self.pis.get('subcompact', \
                     self.pis.get('compact', \
                     self.pis.get('rfcedstyle', 'no'))) == 'yes':
-                    lb = False
+                    leading_blankline = False
                 if style == 'symbols':
                     bullet = self.list_symbols[level % len(self.list_symbols)]
                     bullet += '  '
@@ -171,9 +178,13 @@ class RawTextRfcWriter(BaseRfcWriter):
                     self.list_counters[counter_index] += 1
                     count = self.list_counters[counter_index]
                     bullet = self._format_counter(format_str, count) + ' '
+                if hangIndent:
+                    sub_indent = hangIndent
+                else:
+                    sub_indent = len(bullet)
                 self.write_t_rec(element, bullet=bullet, indent=indent, \
                                  level=level + 1, \
-                                 sub_indent=hangIndent, lb=lb)
+                                 sub_indent=sub_indent, leading_blankline=leading_blankline)
                 t_count += 1
 
         
@@ -197,35 +208,36 @@ class RawTextRfcWriter(BaseRfcWriter):
             indent_scale = 3
         for item in tocindex:
             # Add decoration to counter if it exists, otherwise leave empty
-            counter = ''
-            if item.counter:
-                counter = item.counter + '. '
-                # Extra space on single digit counters
-                if len(item.counter.rsplit('.')[-1]) == 1:
-                    counter += ' '
-            # Get item depth based on its section 'level' attribute
-            depth = item.level - 1
-            if depth < 0 or self.pis.get('tocindent', 'yes') == 'no':
-                depth = 0
-            # Prepend appendix at first level
-            if item.level == 1 and item.appendix:
-                counter = "Appendix " + counter
-            bullet = ' ' * (depth * indent_scale) + counter
-            indent = 3
-            sub_indent = indent + len(bullet)
-            lines = textwrap.wrap(bullet + item.title, 
-                                  self.width - len(str(item.page)) + 1,
-                                  initial_indent=' ' * indent,
-                                  subsequent_indent=' ' * sub_indent)
-            if paging:
-                # Construct dots
-                dots = len(lines[-1]) % 2 and ' ' or '  '
-                dots += '. ' * int((self.width - len(lines[-1]) - len(dots))/2)
-                lines[-1] += dots
-                # Insert page
-                pagestr = ' ' + str(item.page)
-                lines[-1] = lines[-1][:0 - len(pagestr)] + pagestr
-            tmpbuf.extend(lines)
+            if item.level <= tocdepth:
+                counter = ''
+                if item.counter:
+                    counter = item.counter + '. '
+                    # Extra space on single digit counters
+                    if len(item.counter.rsplit('.')[-1]) == 1:
+                        counter += ' '
+                # Get item depth based on its section 'level' attribute
+                depth = item.level - 1
+                if depth < 0 or self.pis.get('tocindent', 'yes') == 'no':
+                    depth = 0
+                # Prepend appendix at first level
+                if item.level == 1 and item.appendix:
+                    counter = "Appendix " + counter
+                bullet = ' ' * (depth * indent_scale) + counter
+                indent = 3
+                sub_indent = indent + len(bullet)
+                pagestr = '%4s' % item.page
+                lines = textwrap.wrap(bullet + item.title, 
+                                      self.width - len(pagestr),
+                                      initial_indent=' ' * indent,
+                                      subsequent_indent=' ' * sub_indent)
+                if paging:
+                    # Construct dots
+                    dots = len(lines[-1]) % 2 and ' ' or ''
+                    dots += ' .' * int((self.width - len(lines[-1]) - len(dots) + 1)/2)
+                    lines[-1] += dots
+                    # Insert page
+                    lines[-1] = lines[-1][:0 - len(pagestr)] + pagestr
+                tmpbuf.extend(lines)
         return tmpbuf
             
     def _write_iref_index(self):
@@ -244,7 +256,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                 alpha_bucket[letter] = [key]
         for letter in sorted(alpha_bucket.keys()):
             # Write letter
-            self._write_text(letter, indent=3, lb=True, buf=tmpbuf)
+            self._write_text(letter, indent=3, leading_blankline=True, buf=tmpbuf)
             for item in alpha_bucket[letter]:
                 pages = self._iref_index[item].pages
                 # Write item
@@ -273,21 +285,21 @@ class RawTextRfcWriter(BaseRfcWriter):
         if xref.text:
             if not target_text.startswith('['):
                 target_text = '(' + target_text + ')'
-            return xref.text + ' ' + target_text
+            return xref.text.rstrip() + ' ' + target_text
         else:
             return target_text
 
-    def _write_ref_element(self, key, text, sub_indent):
+    def _write_ref_element(self, key, text, sub_indent, source_line=None):
         """ Render a single reference element """
         # Use an empty first line if key is too long
         min_spacing = 2
         if len(key) + min_spacing > sub_indent:
-            self._write_text(key, indent=3, lb=True)
-            self._write_text(text, indent=3 + sub_indent)
+            self._write_text(key, indent=3, leading_blankline=True, source_line=source_line)
+            self._write_text(text, indent=3 + sub_indent, source_line=source_line)
         else:
             # Fill space to sub_indent in the bullet
             self._write_text(text, indent=3, bullet=key.ljust(sub_indent), \
-                     sub_indent=sub_indent, lb=True)
+                     sub_indent=sub_indent, leading_blankline=True, source_line=source_line)
     
     def _combine_inline_elements(self, elements):
         """ Shared function for <t> and <c> elements
@@ -348,7 +360,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                 style = element.attrib.get('style', 'emph')
                 edgechar = '?'
                 if style == 'emph':
-                    edgechar = '-'
+                    edgechar = '_'
                 elif style == 'strong':
                     edgechar = '*'
                 elif style == 'verb':
@@ -368,7 +380,15 @@ class RawTextRfcWriter(BaseRfcWriter):
         # Went through all elements, return text with an empty list
         return ''.join(line), []
             
-        
+    def _check_long_lines(self, buf_line, source_line):
+        long_lines = [ (num, line) for num, line in enumerate(self.buf[buf_line:]) if (len(line) > self.width) ]
+        for num, line in long_lines:
+            if source_line:
+                xml2rfc.log.warn("Output line (from source around line %s) is %s characters; longer than %s.  Excess characters: '%s':\n  '%s'\n"
+                    % (source_line+num, len(line), self.width, line[self.width:], line))
+            else:
+                xml2rfc.log.warn("Output line is %s characters; longer than %s.  Excess characters: '%s':\n  '%s'\n"
+                    % (buf_line+num, len(line), self.width, line[self.width:], line))
 
     # ---------------------------------------------------------
     # Base writer overrides
@@ -383,10 +403,10 @@ class RawTextRfcWriter(BaseRfcWriter):
         self.iref_marker = len(self.buf)
 
     def write_raw(self, text, indent=3, align='left', blanklines=0, \
-                  delimiter=None, lb=True):
+                  delimiter=None, leading_blankline=True, source_line=None):
         """ Writes a raw stream of characters, preserving space and breaks """
         if text:
-            if lb:
+            if leading_blankline:
                 # Start with a newline
                 self._lb()
             # Delimiter?
@@ -394,6 +414,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                 self.buf.append(delimiter)
             # Additional blank lines?
             self.buf.extend([''] * blanklines)
+            start_line = len(self.buf)
             # Format the input
             lines = [line.rstrip() for line in text.expandtabs(4).split('\n')]
             # Trim first and last lines if they are blank, whitespace is handled
@@ -426,14 +447,16 @@ class RawTextRfcWriter(BaseRfcWriter):
             # Delimiter?
             if delimiter:
                 self.buf.append(delimiter)
+            if not self.indexmode:
+                self._check_long_lines(start_line, source_line)
 
-    def write_label(self, text, type='figure'):
+    def write_label(self, text, type='figure', source_line=None):
         """ Writes a centered label """
-        self._write_text(text, align='center', lb=True)
+        self._write_text(text, indent=3, align='center', leading_blankline=True, source_line=source_line)
 
-    def write_title(self, title, docName=None):
+    def write_title(self, title, docName=None, source_line=None):
         """ Write the document title and (optional) name """
-        self._write_text(title, lb=True, align='center')
+        self._write_text(title, leading_blankline=True, align='center', source_line=source_line)
         if docName is not None:
             self._write_text(docName, align='center')
 
@@ -442,14 +465,14 @@ class RawTextRfcWriter(BaseRfcWriter):
         """ Write a generic header """
         if bullet:
             bullet += '  '
-        self._write_text(text, bullet=bullet, indent=0, lb=True)
+        self._write_text(text, bullet=bullet, indent=0, leading_blankline=True)
 
     def write_paragraph(self, text, align='left', autoAnchor=None):
-        """ Write a generic paragraph of text """
-        self._write_text(text, indent=3, align=align, lb=True)
+        """ Write a generic paragraph of text.  Used for boilerplate. """
+        self._write_text(text, indent=3, align=align, leading_blankline=True)
 
     def write_t_rec(self, t, indent=3, sub_indent=0, bullet='',
-                     autoAnchor=None, align='left', level=0, lb=True):
+                     autoAnchor=None, align='left', level=0, leading_blankline=True):
         """ Recursively writes a <t> element """
         # Grab any initial text in <t>
         current_text = t.text or ''
@@ -462,9 +485,10 @@ class RawTextRfcWriter(BaseRfcWriter):
             current_text += inline_text
             if (current_text and not current_text.isspace()) or bullet:
                 # Attempt to write a paragraph of inline text
-                self._write_text(current_text, indent=indent, lb=lb, \
-                                sub_indent=sub_indent, bullet=bullet, \
-                                edit=True, align=align)
+                self._write_text(current_text, indent=indent,
+                                leading_blankline=leading_blankline,
+                                sub_indent=sub_indent, bullet=bullet,
+                                edit=True, align=align, source_line=t.sourceline)
             # Clear text
             current_text = ''
 
@@ -481,13 +505,13 @@ class RawTextRfcWriter(BaseRfcWriter):
                     # Call sibling function to construct list
                     self._write_list(element, indent=new_indent, level=level)
                     # Auto-break for tail paragraph
-                    lb = True
+                    leading_blankline = True
                     bullet = ''
 
                 elif element.tag == 'figure':
                     self._write_figure(element)
                     # Auto-break for tail paragraph
-                    lb = True
+                    leading_blankline = True
                     bullet = ''
 
                 elif element.tag == 'vspace':
@@ -495,9 +519,9 @@ class RawTextRfcWriter(BaseRfcWriter):
                     self._vspace(num=int(element.attrib.get('blankLines',
                                          self.defaults['vspace_blanklines'])))
                     # Don't auto-break for tail paragraph
-                    lb = False
+                    leading_blankline = False
                     # Keep indentation
-                    bullet = ' ' * len(bullet)
+                    bullet = ' ' * sub_indent
 
                 # Set tail of element as input text of next paragraph
                 if element.tail:
@@ -506,8 +530,11 @@ class RawTextRfcWriter(BaseRfcWriter):
 
     def write_top(self, left_header, right_header):
         """ Combines left and right lists to write a document heading """
-        # Begin with three blank lines
-        self._lb(num=3)
+        # Begin with a blank line on the first page.  We'll add additional
+        # blank lines at the top later, but those won't be counted as part
+        # of the page linecount.
+        #self._lb(num=3)
+        self._lb()
         heading = []
         for i in range(max(len(left_header), len(right_header))):
             if i < len(left_header):
@@ -520,7 +547,8 @@ class RawTextRfcWriter(BaseRfcWriter):
                 right = ''
             heading.append(xml2rfc.utils.justify_inline(left, '', right, \
                                                         self.width))
-        self.write_raw('\n'.join(heading), align='left', indent=0, lb=False)
+        self.write_raw('\n'.join(heading), align='left', indent=0,
+                        leading_blankline=False, source_line=None)
         # Extra blank line underneath top block
         self._lb()
 
@@ -573,7 +601,7 @@ class RawTextRfcWriter(BaseRfcWriter):
             uri = address.find('uri')
             if uri is not None and uri.text:
                 lines.append('URI:   ' + uri.text)
-        self.write_raw('\n'.join(lines))
+        self.write_raw('\n'.join(lines), indent=self.margin)
         self._lb()
 
     def write_reference_list(self, list):
@@ -581,6 +609,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         refdict = {}
         annotationdict = {}
         refkeys = []
+        refsource = {}
         # [surname, initial.,] "title", (STD), (BCP), (RFC), (Month) Year.
         for i, ref in enumerate(list.findall('reference')):
             refstring = []
@@ -591,13 +620,13 @@ class RawTextRfcWriter(BaseRfcWriter):
             if title is not None and title.text:
                 refstring.append('"' + title.text + '", ')
             else:
-                xml2rfc.log.warn('No title specified in reference', \
+                xml2rfc.log.warn('No title specified in reference',
                                  ref.attrib.get('anchor', ''))
             for seriesInfo in ref.findall('seriesInfo'):
                 if seriesInfo.attrib['name'] == "Internet-Draft":
                     refstring.append(seriesInfo.attrib['value'] + ' (work in progress), ')
                 else:
-                    refstring.append(seriesInfo.attrib['name'] + ' ' + \
+                    refstring.append(seriesInfo.attrib['name'] + ' ' +
                                      seriesInfo.attrib['value'] + ', ')
             date = ref.find('front/date')
             if date is not None:
@@ -618,6 +647,7 @@ class RawTextRfcWriter(BaseRfcWriter):
             else:
                 bullet = '[' + str(i + 1) + ']'
             refdict[bullet] = ''.join(refstring)
+            refsource[bullet] = ref.sourceline
             refkeys.append(bullet)
             # Add annotation if it exists to a separate dict
             if annotation is not None and annotation.text:
@@ -628,11 +658,11 @@ class RawTextRfcWriter(BaseRfcWriter):
         # Hard coded indentation amount
         refindent = 11
         for key in refkeys:
-            self._write_ref_element(key, refdict[key], refindent)
+            self._write_ref_element(key, refdict[key], refindent, source_line=refsource[key])
             # Render annotation as a separate paragraph
             if key in annotationdict:
-                self._write_text(annotationdict[key], indent=refindent + 3, \
-                                 lb=True)
+                self._write_text(annotationdict[key], indent=refindent + 3,
+                                 leading_blankline=True, source_line=refsource[key])
 
     def draw_table(self, table, table_num=None):
         # First construct a 2d matrix from the table
@@ -806,13 +836,17 @@ class RawTextRfcWriter(BaseRfcWriter):
 
         # Finally, write the table to the buffer with proper alignment
         align = table.attrib.get('align', 'center')
-        self.write_raw('\n'.join(output), align=align)
+        self.write_raw('\n'.join(output), align=align, indent=self.margin,
+                        source_line=table.sourceline)
 
     def insert_anchor(self, text):
         # No anchors for text
         pass
 
-    def pre_processing(self):
+    def pre_indexing(self):
+        pass
+
+    def pre_rendering(self):
         # Discard buffer from indexing pass
         self.buf = []
         
@@ -821,7 +855,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         self.edit_counter = 0   # Counter for edit marks
         self.eref_counter = 0   # Counter for <eref> elements
 
-    def post_processing(self):
+    def post_rendering(self):
         # Insert the TOC and IREF into the main buffer
         self.output = self.buf[:self.toc_marker] + \
                       self._write_toc() + \
@@ -831,6 +865,11 @@ class RawTextRfcWriter(BaseRfcWriter):
 
     def write_to_file(self, file):
         """ Writes the buffer to the specified file """
+        # write initial blank lines, not counted against page size...
+        if self.draft:
+            file.write(os.linesep*3)
+        else:
+            file.write(os.linesep*6)
         for line in self.output:
-            file.write(line)
+            file.write(line.rstrip(" \t"))
             file.write(os.linesep)
